import os
from yaclog.changelog import Changelog


def read(path):
    """
    Create a new Changelog object from the given path
    :param path: a path to a markdown changelog file
    :return: a parsed Changelog object
    """
    return Changelog(path)



"""
Contains the `Changelog` class that represents a parsed changelog file that can be read from and written to
disk as markdown, as well as the `VersionEntry` class that represents a single version within that changelog.
"""

#  yaclog: yet another changelog tool
#  Copyright (c) 2021. Andrew Cassidy
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import datetime
import os
import re
from typing import List, Optional, Dict

import click  # only for styling

import yaclog.markdown as markdown
import yaclog.version


class VersionEntry:
    """
    A serialized representation of a single version entry in a `Changelog`,
    containing the changes made since the previous version
    """

    _header_regex = re.compile(  # THE LANGUAGE OF THE GODS
        r"##\s+(?P<name>.*?)(?:\s+-)?(?:\s+(?P<date>\d{4}-\d{2}-\d{2}))?(?P<tags>(?:\s+\[[^]]*?])*)\s*$")

    _tag_regex = re.compile(r'\[(?P<tag>[^]]*?)]')

    def __init__(self, name: str = 'Unreleased',
                 date: Optional[datetime.date] = None, tags: Optional[List[str]] = None,
                 link: Optional[str] = None, link_id: Optional[str] = None, line_no: Optional[int] = None):
        """
        :param str name: The version's name
        :param Optional[datetime.date] date: When the version was released
        :param tags: The version's tags
        :param link: The version's URL
        :param link_id: The version's link ID
        :param line_no: What line in the original file the version starts on
        """

        self.name: str = name
        """The version's name"""

        self.date: Optional[datetime.date] = date
        """When the version was released"""

        self.tags: List[str] = tags if tags else []
        """The version's tags"""

        self.link: Optional[str] = link
        """The version's URL"""

        self.link_id: Optional[str] = link_id
        """The version's link ID, uses the version name by default when writing"""

        self.line_no: Optional[int] = line_no
        """What line the version occurs at in the file, or `None` if the version was not read from a file. 
        This is not guaranteed to be correct after the changelog has been modified, 
        and it has no effect on the written file"""

        self.sections: Dict[str, List[str]] = {'': []}
        """The dictionary of change entries in the version, organized by section. 
        Uncategorized changes have a section of an empty string."""

    @classmethod
    def from_header(cls, header: str, line_no: Optional[int] = None) -> VersionEntry:
        """
        Create a new version entry from a markdown header

        :param header: A markdown header to parse
        :param line_no: Line number the header is on
        :return: a new VersionEntry with the header's information
        """
        version = cls(line_no=line_no)

        match = cls._header_regex.match(header)
        assert match, f'failed to parse version header: "{header}"'

        version.name, version.link, version.link_id = markdown.strip_link(match['name'])

        if match['date']:
            try:
                version.date = datetime.date.fromisoformat(match['date'])
            except ValueError:
                return cls(name=header.lstrip('#').strip(), line_no=line_no)

        if match['tags']:
            version.tags = [m['tag'].upper() for m in cls._tag_regex.finditer(match['tags'])]

        return version

    def add_entry(self, contents: str, section: str = '') -> None:
        """
        Add a new entry to the version

        :param contents: The contents string to add
        :param section: Which section to add to.
        """

        section = section.title()
        if section not in self.sections.keys():
            self.sections[section] = []

        self.sections[section].append(contents)

    def body(self, md: bool = True, color: bool = False) -> str:
        """
        Get the version's body as a string

        :param md: Format headings as markdown
        :param color: Add color codes to the string for display in a terminal
        :return: The formatted version body, without the version header
        """

        segments = []

        for section, entries in self.sections.items():
            if section:
                if md:
                    prefix = '### '
                    title = section.title()
                else:
                    prefix = ''
                    title = section.upper()

                if color:
                    prefix = click.style(prefix, fg='bright_black')
                    title = click.style(title, fg='cyan', bold=True)

                segments.append(prefix + title)

            if len(entries) > 0:
                segments += entries

        return markdown.join(segments)

    def header(self, md: bool = True, color: bool = False) -> str:
        """
        Get the version's header as a string

        :param md: Format headings as markdown
        :param color: Add color codes to the string for display in a terminal
        :return: The formatted version header
        """

        if md:
            prefix = '## '
        else:
            prefix = ''

        segments = []

        if self.link and md:
            segments.append(f'[{self.name}]')
        else:
            segments.append(self.name)

        if self.date or len(self.tags) > 0:
            segments.append('-')

        if self.date:
            segments.append(self.date.isoformat())

        segments += [f'[{t.upper()}]' for t in self.tags]

        title = ' '.join(segments)

        if color:
            prefix = click.style(prefix, fg='bright_black')
            title = click.style(title, fg='blue', bold=True)

        return prefix + title

    def text(self, md: bool = True, color: bool = False) -> str:
        """
        Get the version's contents as a string

        :param md: Format headings as markdown
        :param color: Add color codes to the string for display in a terminal
        :return: The formatted version header and body
        """

        contents = self.header(md, color)
        body = self.body(md, color)
        if body:
            contents += '\n\n' + body
        return contents

    @property
    def released(self) -> bool:
        """Returns true if a PEP440 version number is present in the version name, and has no prerelease segments"""
        return yaclog.version.is_release(self.name)

    @property
    def version(self):
        """Returns the PEP440 version number from the version name, or `None` if none is found"""
        return yaclog.version.extract_version(self.name)[0]

    def __str__(self) -> str:
        return self.header(False)


class Changelog:
    """
    A serialized representation of a Markdown changelog made up of a preamble, multiple versions, and a link table.
    """

    def __init__(self, path=None,
                 preamble: str = "# Changelog\n\nAll notable changes to this project will be documented in this file"):
        """
        Contents will be automatically read from disk if the file exists

        :param path: The changelog's path on disk.
        :param str preamble: The changelog preamble to use if the file does not exist.
        """
        self.path = os.path.abspath(path) if path else None
        """The path of the changelog's file on disk"""

        self.preamble: str = preamble
        """Any text at the top of the changelog before any version information, including the title.
        It can contain the title, an explanation of the file's purpose, as well as any general machine-readable 
        information for use with other tools."""

        self.versions: List[VersionEntry] = []
        """A list of versions in the changelog, with the most recent version first"""

        self.links: Dict[str, str] = {}
        """Link definitions at the end of the changelog, as a dictionary of ``{id: url}``"""

        if path and os.path.exists(path):
            self.read()

    def read(self, path=None) -> None:
        """
        Read a markdown changelog file from disk. The object's contents will be overwritten by the file contents if
        reading is successful.

        :param path: The changelog's path on disk. By default, :py:attr:`~Changelog.path` is used
        """

        if not path:
            # use the object path if none was provided
            path = self.path

        # Read file
        with open(path, 'r') as fp:
            tokens, links = markdown.tokenize(fp.read())

        section = ''
        versions = []
        preamble_segments = []

        for token in tokens:
            text = '\n'.join(token.lines)

            if token.kind == 'h2':
                # start of a version
                versions.append(VersionEntry.from_header(text, line_no=token.line_no))
                section = ''

            elif len(versions) == 0:
                # we haven't encountered any version headers yet,
                # so its best to just add this line to the preamble
                preamble_segments.append(text)

            elif token.kind == 'h3':
                # start of a version section
                section = text.strip('#').strip()
                if section not in versions[-1].sections.keys():
                    versions[-1].sections[section] = []

            else:
                # change log entry
                versions[-1].sections[section].append(text)

        # handle links
        for version in versions:
            if match := re.fullmatch(r'\[(.*)]', version.name):
                # ref-matched link
                link_id = match[1].lower()
                if link_id in links:
                    version.link = links[link_id]
                    version.link_id = None
                    version.name = match[1]

            elif version.link_id in links:
                # id-matched link
                version.link = links[version.link_id]

        self.preamble = markdown.join(preamble_segments)
        self.versions = versions
        self.links = links

    def write(self, path=None) -> None:
        """
        Write a changelog to a Markdown file.

        :param path: The changelog's path on disk. By default, :py:attr:`~Changelog.path` is used.
        """

        if path is None:
            # use the object path if none was provided
            path = self.path

        segments = []

        if self.preamble:
            segments.append(self.preamble)

        v_links = {**self.links}

        for version in self.versions:
            if version.link:
                v_links[version.name.lower()] = version.link

            segments.append(version.text() + '\n')

        segments += [f'[{link_id}]: {link}' for link_id, link in v_links.items()]

        text = markdown.join(segments)

        with open(path, 'w') as fp:
            fp.write(text)

    def add_version(self, index: int = 0, *args, **kwargs) -> VersionEntry:
        """
        Add a new version to the changelog

        :param index: Where to add the new version in the log. Defaults to the top
        :param args: args to forward to the :py:class:`VersionEntry` constructor
        :param kwargs: kwargs to forward to the :py:class:`VersionEntry` constructor
        :return: The created entry
        """
        self.versions.insert(index, version := VersionEntry(*args, **kwargs))
        return version

    def current_version(self, released: Optional[bool] = None, new_version: bool = False,
                        new_version_name: str = 'Unreleased') -> VersionEntry:
        """
        Get the current version from the changelog

        :param released: if the returned version should be a released version,
            an unreleased version, or `None` to return the most recent
        :param new_version: if a new version should be created if none exist.
        :param new_version_name: The name of the version to create if there
            are no matches and ``new_version`` is True.
        :return: The current version matching the criteria,
            or `None` if ``new_version`` is disabled and none are found.
        """

        # return the first version that matches `released`
        for version in self.versions:
            if version.released == released or released is None:
                return version

        # fallback if none are found
        if new_version:
            return self.add_version(name=new_version_name)
        else:
            if released is not None:
                raise ValueError(f'Changelog has no current version matching released={released}')
            else:
                raise ValueError('Changelog has no current version')

    def get_version(self, name: Optional[str] = None) -> VersionEntry:
        """
        Get a version from the changelog by name.

        :param name: The name of the version to get, or `None` to return the most recent.
            The first version with this value in its name is returned.
        :return: The first version with the selected name
        """

        for version in self.versions:
            if name in version.name or name is None:
                return version
        raise KeyError(f'Version {name} not found in changelog')

    def __getitem__(self, item: str) -> VersionEntry:
        return self.get_version(item)

    def __len__(self) -> int:
        return len(self.versions)



#  yaclog: yet another changelog tool
#  Copyright (c) 2021. Andrew Cassidy
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import os.path

import click
import git

import yaclog.version
from yaclog.changelog import Changelog


@click.group()
@click.option('--path', envvar='YACLOG_PATH', metavar='FILE', default='CHANGELOG.md', show_default=True,
              type=click.Path(dir_okay=False, writable=True, readable=True),
              help='Location of the changelog file.')
@click.version_option()
@click.pass_context
def cli(ctx, path):
    """Manipulate markdown changelog files."""
    if not (ctx.invoked_subcommand == 'init') and not os.path.exists(path):
        # file does not exist and this isn't the init command
        raise click.FileError(f'Changelog file {path} does not exist. Create it by running yaclog init.')

    ctx.obj = yaclog.read(path)


@cli.command()
@click.pass_obj
def init(obj: Changelog):
    """Create a new changelog file."""
    if os.path.exists(obj.path):
        click.confirm(f'Changelog file {obj.path} already exists. Would you like to overwrite it?', abort=True)
        os.remove(obj.path)

    yaclog.Changelog(obj.path).write()
    click.echo(f'Created new changelog file at {obj.path}')


@cli.command('format')  # dont accidentally hide the `format` python builtin
@click.pass_obj
def reformat(obj: Changelog):
    """Reformat the changelog file."""
    obj.write()
    click.echo(f'Reformatted changelog file at {obj.path}')


# noinspection PyShadowingNames
@cli.command(short_help='Show changes from the changelog file')
@click.option('--all', '-a', 'all_versions', is_flag=True, help='Show the entire changelog.')
@click.option('--markdown/--txt', '-m/-t', default=False, help='Display as markdown or plain text.')
@click.option('--full', '-f', 'mode', flag_value='full', default=True,
              help='Show version header and body.')
@click.option('--name', '-n', 'mode', flag_value='name',
              help='Show only the version name')
@click.option('--body', '-b', 'mode', flag_value='body',
              help='Show only the version body.')
@click.option('--header', '-h', 'mode', flag_value='header',
              help='Show only the version header.')
@click.argument('version_names', metavar='VERSIONS', type=str, nargs=-1)
@click.pass_obj
def show(obj: Changelog, all_versions, markdown, mode, version_names):
    """
    Show the changes for VERSIONS.

    VERSIONS is a list of versions to print. If not given, the most recent version is used.
    """

    functions = {
        'full': (lambda v, k: v.text(**k)),
        'name': (lambda v, k: v.name),
        'body': (lambda v, k: v.body(**k)),
        'header': (lambda v, k: v.header(**k)),
    }

    str_func = functions[mode]

    try:
        if all_versions:
            versions = obj.versions
        elif len(version_names) == 0:
            versions = [obj.current_version()]
        else:
            versions = [obj.get_version(name) for name in version_names]
    except KeyError as k:
        raise click.BadArgumentUsage(str(k))
    except ValueError as v:
        raise click.ClickException(str(v))

    kwargs = {'md': markdown, 'color': True}

    sep = '\n\n' if mode == 'body' or mode == 'full' else '\n'
    click.echo(sep.join([str_func(v, kwargs) for v in versions]))


@cli.command(short_help='Modify version tags')
@click.option('--add/--delete', '-a/-d', default=True, is_flag=True, help='Add or delete tags')
@click.argument('tag_name', metavar='TAG', type=str)
@click.argument('version_name', metavar='VERSION', type=str, required=False)
@click.pass_obj
def tag(obj: Changelog, add, tag_name: str, version_name: str):
    """
    Modify TAG on VERSION.

    VERSION is the name of a version to add tags to. If not given, the most recent version is used.
    """
    tag_name = tag_name.upper()
    try:
        if version_name:
            version = obj.get_version(version_name)
        else:
            version = obj.current_version()
    except KeyError as k:
        raise click.BadArgumentUsage(str(k))
    except ValueError as v:
        raise click.ClickException(str(v))

    if add:
        version.tags.append(tag_name)
    else:
        try:
            version.tags.remove(tag_name)
        except ValueError:
            raise click.BadArgumentUsage(f"Tag {tag_name} not found in version {version.name}.")

    obj.write()


@cli.command(short_help='Add entries to the changelog.')
@click.option('--bullet', '-b', 'bullets', metavar='TEXT', multiple=True, type=str, help='Add a bullet point.')
@click.option('--paragraph', '-p', 'paragraphs', metavar='TEXT', multiple=True, type=str, help='Add a paragraph')
@click.argument('section_name', metavar='SECTION', type=str, default='', required=False)
@click.argument('version_name', metavar='VERSION', type=str, default=None, required=False)
@click.pass_obj
def entry(obj: Changelog, bullets, paragraphs, section_name, version_name):
    """
    Add entries to SECTION in VERSION

    SECTION is the name of the section to append to. If not given, entries will be uncategorized.

    VERSION is the name of the version to append to. If not given, the most recent version will be used,
    or a new 'Unreleased' version will be added if the most recent version has been released.
    """

    section_name = section_name.title()
    try:
        if version_name:
            version = obj.get_version(version_name)
        else:
            version = obj.current_version(released=False, new_version=True)
    except KeyError as k:
        raise click.BadArgumentUsage(str(k))

    for p in paragraphs:
        version.add_entry(p, section_name)

    for b in bullets:
        version.add_entry('- ' + b, section_name)

    obj.write()
    count = len(paragraphs) + len(bullets)
    message = f"Created {count} {['entry', 'entries'][min(count - 1, 1)]}"
    if section_name:
        message += f" in section {click.style(section_name, fg='cyan')}"
    if version.name.lower() != 'unreleased':
        message += f" in version {click.style(version.name, fg='blue')}"
    click.echo(message)


@cli.command(short_help='Release versions.')
@click.option('-M', '--major', 'rel_seg', flag_value=0, type=int, default=None,
              help='Increment major version number.')
@click.option('-m', '--minor', 'rel_seg', flag_value=1, type=int,
              help='Increment minor version number.')
@click.option('-p', '--patch', 'rel_seg', flag_value=2, type=int,
              help='Increment patch number.')
@click.option('-a', '--alpha', 'pre_seg', flag_value='a', type=str, default=None,
              help='Increment alpha version number.')
@click.option('-b', '--beta', 'pre_seg', flag_value='b', type=str,
              help='Increment beta version number.')
@click.option('-r', '--rc', 'pre_seg', flag_value='rc', type=str,
              help='Increment release candidate version number.')
@click.option('-f', '--full', 'pre_seg', flag_value='',
              help='Clear the prerelease value creating a full release.')
@click.option('-c', '--commit', is_flag=True,
              help='Create a git commit tagged with the new version number. '
                   'If there are no changes to commit, the current commit will be tagged instead.')
@click.argument('version_name', metavar='VERSION', type=str, default=None, required=False)
@click.pass_obj
def release(obj: Changelog, version_name, rel_seg, pre_seg, commit):
    """
    Release VERSION, or a version incremented from the last release.

    VERSION is the name of the version to release. If VERSION is not provided but increment options are, then the most
    recent valid PEP440 version number is used instead.

    The most recent version in the log will be renamed (except by the --commit option) by using the VERSION as well as
    any increment options. Increment options will always reset the later segments, and prerelease increments will clear
    other kinds of prerelease.
    """

    if rel_seg is None and pre_seg is None and not version_name and not commit:
        click.echo('Nothing to release!')
        raise click.Abort

    cur_version = obj.current_version()
    old_name = cur_version.name

    if version_name:
        new_name = version_name
    else:
        for v in obj.versions:
            if v.version is not None:
                new_name = v.name
                break
        else:
            new_name = '0.0.0'

    if rel_seg is not None or pre_seg is not None:
        new_name = yaclog.version.increment_version(new_name, rel_seg, pre_seg)

    if new_name != old_name:
        if yaclog.version.is_release(old_name):
            click.confirm(
                f"Rename release version {click.style(old_name, fg='blue')} "
                f"to {click.style(new_name, fg='blue')}?",
                abort=True)

        cur_version.name = new_name
        cur_version.date = datetime.datetime.utcnow().date()

        obj.write()
        click.echo(f"Renamed {click.style(old_name, fg='blue')} to {click.style(new_name, fg='blue')}")

    if commit:
        repo = git.Repo(os.curdir)

        if repo.bare:
            raise click.BadOptionUsage('commit', f'Directory {os.path.abspath(os.curdir)} is not a git repo')

        repo.index.add(obj.path)

        tracked = len(repo.index.diff(repo.head.commit))
        untracked = len(repo.index.diff(None))

        message = [['Create tag', 'Commit and create tag'][min(tracked, 1)], 'for']

        if not cur_version.released:
            message.append('non-release')

        message.append(f"version {click.style(new_name, fg='blue')}?")

        if untracked > 0:
            message.append(click.style(
                f"You have {untracked} untracked file{'s'[:untracked]} that will not be included!",
                fg='red', bold=True))

        click.confirm(' '.join(message), abort=True)

        if tracked > 0:
            commit = repo.index.commit(f'Release {cur_version.name}\n\n{cur_version.body()}')
            click.echo(f"Created commit {click.style(repo.head.commit.hexsha[0:7], fg='green')}")
        else:
            commit = repo.head.commit

        short_version, *_ = yaclog.version.extract_version(cur_version.name)
        if not short_version:
            short_version = cur_version.name.replace(' ', '-')

        repo_tag = repo.create_tag(short_version, ref=commit, message=cur_version.body(False))
        click.echo(f"Created tag {click.style(repo_tag.name, fg='green')}.")


if __name__ == '__main__':
    cli()



"""
Various helper functions for analyzing and manipulating :pep:`440` version numbers,
meant to augment the `packaging.version` module.
"""

#  yaclog: yet another changelog tool
#  Copyright (c) 2021. Andrew Cassidy
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
from typing import Optional, Tuple

from packaging.version import Version, VERSION_PATTERN

version_regex = re.compile(VERSION_PATTERN, re.VERBOSE | re.IGNORECASE)


def extract_version(version_str: str) -> Tuple[Optional[Version], int, int]:
    """
    Extracts a :pep:`440` version object from a string which may have other text

    :param version_str: The input string to extract from
    :return: A tuple of (version, start, end), where start and end are the span of the version in the original string
    """
    match = version_regex.search(version_str)
    if not match:
        return None, -1, -1
    return (Version(match[0]),) + match.span()


def increment_version(version_str: str, rel_seg: int = None, pre_seg: str = None) -> str:
    """
    Increment the :pep:`440` version number in a string

    :param version_str: The input string to manipulate
    :param rel_seg: Which segment of the "release" value to increment, if any
    :param pre_seg: Which kind of prerelease to use, if any. An empty string clears the prerelease field.
    :return: The original string with the version number incremented
    """
    v, *span = extract_version(version_str)
    epoch = v.epoch
    release = v.release
    pre = v.pre
    post = v.post
    dev = v.dev
    local = v.local

    if rel_seg is not None:
        if len(release) <= rel_seg:
            release += (0,) * (1 + rel_seg - len(release))
        release = release[0:rel_seg] + (release[rel_seg] + 1,) + (0,) * (len(release) - rel_seg - 1)
        pre = None

    if pre_seg is not None:
        if pre_seg == '':  # full release, clear prerelease field
            pre = None
        elif pre and pre[0] == pre_seg:  # increment current prerelease type
            pre = (pre_seg, pre[1] + 1)
        else:
            pre = (pre_seg, 1)  # set prerelease field

    new_v = join_version(epoch, release, pre, post, dev, local)
    return version_str[0:span[0]] + new_v + version_str[span[1]:]


def join_version(epoch, release, pre, post, dev, local) -> str:
    """Join multiple segments of a :pep:`440` version"""
    parts = []

    # Epoch
    if epoch != 0:
        parts.append(f"{epoch}!")

    # Release segment
    parts.append(".".join(str(x) for x in release))

    # Pre-release
    if pre is not None:
        parts.append("".join(str(x) for x in pre))

    # Post-release
    if post is not None:
        parts.append(f".post{post}")

    # Development release
    if dev is not None:
        parts.append(f".dev{dev}")

    # Local version segment
    if local is not None:
        parts.append(f"+{local}")

    return "".join(parts)


def is_release(version_str: str) -> bool:
    """
    Check if a version string is a release version

    :param version_str: the input string to check
    :return: True if the input contains a released :pep:`440` version,
        or False if a prerelease version or no version is found
    """
    v, *span = extract_version(version_str)
    if v:
        return not (v.is_devrelease or v.is_prerelease)
    else:
        return False
